@@ -19,10 +19,11 @@
 #define ESMTP_SIZE	0x02
 #define ESMTP_ETRN	0x04
 #define ESMTP_ATRN	0x08		/* used with ODMR, RFC 2645 */
+#define ESMTP_AUTH	0x10
 
 void SMTP_setmode(char);
 int SMTP_helo(int socket,const char *host);
-int SMTP_ehlo(int socket,const char *host,int *opt);
+int SMTP_ehlo(int socket,const char *host, char *name, char *passwd, int *opt);
 int SMTP_from(int socket,const char *from,const char *opts);
 int SMTP_rcpt(int socket,const char *to);
 int SMTP_data(int socket);
